@@ -9,6 +9,8 @@ NAME_RE = r"[a-zA-Z][-a-zA-Z0-9_]*"
 _interp_regex = re.compile(r'(?<!\$)(\$(?:(%(n)s)|{(%(n)s)}))'
     % ({'n': NAME_RE}))
 
+CONTEXT_MASK = text_type('%s\x04%s')
+
 class TranslationString(text_type):
     """
     The constructor for a :term:`translation string`.  A translation
@@ -30,7 +32,7 @@ class TranslationString(text_type):
     ``default`` represents an explicit *default text* for this
     translation string.  Default text appears when the translation
     string cannot be translated.  Usually, the ``msgid`` of a
-    translation string serves double duty as as its default text.
+    translation string serves double duty as its default text.
     However, using this option you can provide a different default
     text for this translation string.  This feature is useful when the
     default of a translation string is too complicated or too long to
@@ -45,6 +47,9 @@ class TranslationString(text_type):
     string` *replacement marker* instances found within the ``msgid``
     (or ``default``) value of this translation string.
 
+    ``context`` represents the :term:`translation context`.  By default,
+    the translation context is ``None``.
+
     After a translation string is constructed, it behaves like most
     other ``unicode`` objects; its ``msgid`` value will be displayed
     when it is treated like a ``unicode`` object.  Only when its
@@ -56,9 +61,9 @@ class TranslationString(text_type):
     ``mapping`` attribute.  The object otherwise behaves much like a
     Unicode string.
     """
-    __slots__ = ('domain', 'default', 'mapping')
+    __slots__ = ('domain', 'context', 'default', 'mapping')
 
-    def __new__(self, msgid, domain=None, default=None, mapping=None):
+    def __new__(self, msgid, domain=None, default=None, mapping=None, context=None):
 
         # NB: this function should never never lose the *original
         # identity* of a non-``None`` but empty ``default`` value
@@ -67,20 +72,39 @@ class TranslationString(text_type):
         self = text_type.__new__(self, msgid)
         if isinstance(msgid, self.__class__):
             domain = domain or msgid.domain and msgid.domain[:]
+            context = context or msgid.context and msgid.context[:]
             default = default or msgid.default and msgid.default[:]
             mapping = mapping or msgid.mapping and msgid.mapping.copy()
             msgid = text_type(msgid)
         self.domain = domain
+        self.context = context
         if default is None:
             default = text_type(msgid)
         self.default = default
         self.mapping = mapping
         return self
 
+    def __mod__(self, options):
+        """Create a new TranslationString instance with an updated mapping.
+        This makes it possible to use the standard python %-style string
+        formatting with translatable strings. Only dictionary
+        arguments are supported.
+        """
+        if not isinstance(options, dict):
+            raise ValueError(
+                    'Can only interpolate translationstring '
+                    'with dictionaries.')
+        if self.mapping:
+            mapping = self.mapping.copy()
+            mapping.update(options)
+        else:
+            mapping = options.copy()
+        return TranslationString(self, mapping=mapping)
+
     def interpolate(self, translated=None):
         """ Interpolate the value ``translated`` which is assumed to
         be a Unicode object containing zero or more *replacement
-        markers* (``${foo}`` or ``${bar}``) using the ``mapping``
+        markers* (``$foo`` or ``${bar}``) using the ``mapping``
         dictionary attached to this instance.  If the ``mapping``
         dictionary is empty or ``None``, no interpolation is
         performed.
@@ -94,7 +118,7 @@ class TranslationString(text_type):
         # NB: this function should never never lose the *original
         # identity* of a non-``None`` but empty ``default`` value it
         # is provided.  If (translated == default) , it should return the
-        # *orignal* default, not a derivation.  See the comment below in
+        # *original* default, not a derivation.  See the comment below in
         # ChameleonTranslate.
 
         if self.mapping and translated:
@@ -109,9 +133,9 @@ class TranslationString(text_type):
         return self.__class__, self.__getstate__()
 
     def __getstate__(self):
-        return text_type(self), self.domain, self.default, self.mapping
+        return text_type(self), self.domain, self.default, self.mapping, self.context
 
-def TranslationStringFactory(domain):
+def TranslationStringFactory(factory_domain):
     """ Create a factory which will generate translation strings
     without requiring that each call to the factory be passed a
     ``domain`` value.  A single argument is passed to this class'
@@ -122,12 +146,20 @@ def TranslationStringFactory(domain):
     ``__call__`` method of an instance of this class have the meaning
     as described by the constructor of the
     :class:`translationstring.TranslationString`"""
-    def create(msgid, mapping=None, default=None):
+    def create(msgid, mapping=None, default=None, context=None):
         """ Provided a msgid (Unicode object or :term:`translation
         string`) and optionally a mapping object, and a *default
         value*, return a :term:`translation string` object."""
+
+        # if we are passing in a TranslationString as the msgid, then
+        # use its domain
+        if isinstance(msgid, TranslationString):
+            domain = msgid.domain or factory_domain
+        else:
+            domain = factory_domain
+
         return TranslationString(msgid, domain=domain, default=default,
-                                 mapping=mapping)
+                                 mapping=mapping, context=context)
     return create
 
 def ChameleonTranslate(translator):
@@ -178,13 +210,14 @@ def ChameleonTranslate(translator):
         # spray these indignant comments all over this module. ;-)
 
         if not isinstance(msgid, string_types):
+            if msgid is not None:
+                msgid = text_type(msgid)
             return msgid
 
         tstring = msgid
 
         if not hasattr(tstring, 'interpolate'):
-            tstring = TranslationString(msgid, domain, default, mapping)
-
+            tstring = TranslationString(msgid, domain, default, mapping, context)
         if translator is None:
             result = tstring.interpolate()
         else:
@@ -194,33 +227,48 @@ def ChameleonTranslate(translator):
 
     return translate
 
-def ugettext_policy(translations, tstring, domain):
+def ugettext_policy(translations, tstring, domain, context):
     """ A translator policy function which unconditionally uses the
     ``ugettext`` API on the translations object."""
-    
+
     if PY3: # pragma: no cover
         _gettext = translations.gettext
     else: # pragma: no cover
         _gettext = translations.ugettext
-    
-    return _gettext(tstring)
 
-def dugettext_policy(translations, tstring, domain):
+    if context:
+	# Workaround for http://bugs.python.org/issue2504?
+        msgid = CONTEXT_MASK % (context, tstring)
+    else:
+        msgid = tstring
+
+    translated = _gettext(msgid)
+    return tstring if translated == msgid else translated
+
+def dugettext_policy(translations, tstring, domain, context):
     """ A translator policy function which assumes the use of a
     :class:`babel.support.Translations` translations object, which
     supports the dugettext API; fall back to ugettext."""
     if domain is None:
         default_domain = getattr(translations, 'domain', None) or 'messages'
         domain = getattr(tstring, 'domain', None) or default_domain
+    context = context or getattr(tstring, 'context', None)
+    if context:
+	# Workaround for http://bugs.python.org/issue2504?
+        msgid = CONTEXT_MASK % (context, tstring)
+    else:
+        msgid = tstring
+
     if getattr(translations, 'dugettext', None) is not None:
-        return translations.dugettext(domain, tstring)
-    
-    if PY3: # pragma: no cover
-        _gettext = translations.gettext
-    else: # pragma: no cover
-        _gettext = translations.ugettext
-    
-    return _gettext(tstring)
+        translated = translations.dugettext(domain, msgid)
+    else:
+        if PY3: # pragma: no cover
+            _gettext = translations.gettext
+        else: # pragma: no cover
+            _gettext = translations.ugettext
+
+        translated = _gettext(msgid)
+    return tstring if translated == msgid else translated
 
 def Translator(translations=None, policy=None):
     """
@@ -233,7 +281,7 @@ def Translator(translations=None, policy=None):
     ``policy`` should be a callable which accepts three arguments:
     ``translations``, ``tstring`` and ``domain``.  It must perform the
     actual translation lookup.  If ``policy`` is ``None``, the
-    :func:`translationstring.dugettext_policy` policy will be used. 
+    :func:`translationstring.dugettext_policy` policy will be used.
 
     The callable returned accepts three arguments: ``tstring``
     (required), ``domain`` (optional) and ``mapping`` (optional).
@@ -250,12 +298,21 @@ def Translator(translations=None, policy=None):
     """
     if policy is None:
         policy = dugettext_policy
-    def translator(tstring, domain=None, mapping=None):
+    def translator(tstring, domain=None, mapping=None, context=None):
         if not hasattr(tstring, 'interpolate'):
-            tstring = TranslationString(tstring, domain=domain, mapping=mapping)
+            tstring = TranslationString(tstring, domain=domain, mapping=mapping, context=context)
+        elif mapping:
+            if tstring.mapping:
+                new_mapping = tstring.mapping.copy()
+                new_mapping.update(mapping)
+            else:
+                new_mapping = mapping
+            tstring = TranslationString(tstring, domain=domain, mapping=new_mapping, context=context)
         translated = tstring
+        domain = domain or tstring.domain
+        context = context or tstring.context
         if translations is not None:
-            translated = policy(translations, tstring, domain)
+            translated = policy(translations, tstring, domain, context)
         if translated == tstring:
             translated = tstring.default
         if translated and '$' in translated and tstring.mapping:
@@ -263,34 +320,46 @@ def Translator(translations=None, policy=None):
         return translated
     return translator
 
-def ungettext_policy(translations, singular, plural, n, domain):
+def ungettext_policy(translations, singular, plural, n, domain, context):
     """ A pluralizer policy function which unconditionally uses the
     ``ungettext`` API on the translations object."""
-        
+
     if PY3: # pragma: no cover
         _gettext = translations.ngettext
     else: # pragma: no cover
         _gettext = translations.ungettext
-    
-    return _gettext(singular, plural, n)
 
-def dungettext_policy(translations, singular, plural, n, domain):
+    if context:
+	# Workaround for http://bugs.python.org/issue2504?
+        msgid = CONTEXT_MASK % (context, singular)
+    else:
+        msgid = singular
+
+    translated = _gettext(msgid, plural, n)
+    return singular if translated == msgid else translated
+
+def dungettext_policy(translations, singular, plural, n, domain, context):
     """ A pluralizer policy function which assumes the use of the
     :class:`babel.support.Translations` class, which supports the
     dungettext API; falls back to ungettext."""
-    
+
     default_domain = getattr(translations, 'domain', None) or 'messages'
     domain = domain or default_domain
-    
+    if context:
+	# Workaround for http://bugs.python.org/issue2504?
+        msgid = CONTEXT_MASK % (context, singular)
+    else:
+        msgid = singular
     if getattr(translations, 'dungettext', None) is not None:
-        return translations.dungettext(domain, singular, plural, n)
-        
-    if PY3: # pragma: no cover
-        _gettext = translations.ngettext
-    else: # pragma: no cover
-        _gettext = translations.ungettext
-    
-    return _gettext(singular, plural, n)
+        translated = translations.dungettext(domain, msgid, plural, n)
+    else:
+        if PY3: # pragma: no cover
+            _gettext = translations.ngettext
+        else: # pragma: no cover
+            _gettext = translations.ungettext
+
+        translated = _gettext(msgid, plural, n)
+    return singular if translated == msgid else translated
 
 def Pluralizer(translations=None, policy=None):
     """
@@ -326,10 +395,10 @@ def Pluralizer(translations=None, policy=None):
         policy = dungettext_policy
     if translations is None:
         translations = NullTranslations()
-    def pluralizer(singular, plural, n, domain=None, mapping=None):
+    def pluralizer(singular, plural, n, domain=None, mapping=None, context=None):
         """ Pluralize this object """
         translated = text_type(
-            policy(translations, singular, plural, n, domain))
+            policy(translations, singular, plural, n, domain, context))
         if translated and '$' in translated and mapping:
             return TranslationString(translated, mapping=mapping).interpolate()
         return translated
